#ifndef _IDE_I_H
#define _IDE_I_H

/* This file is ripped from ide.c. It has been modified to allow the
   Ultra66 driver module to include various functions conatined within
   it. This is necessary because many symbols that we would like to use
   have not been exported for use by modules. Most functions that are
   normally part of the kernel have been renamed by suffixing an "_i" to
   avoid confusion. -frankt
*/

#include "ide.h"

ide_settings_t *ide_find_setting_by_name(ide_drive_t *drive, char *name)
{
	ide_settings_t *setting = drive->settings;

	while (setting) {
		if (strcmp(setting->name, name) == 0)
			break;
		setting = setting->next;
	}
	return setting;
}

int ide_write_setting(ide_drive_t *drive, ide_settings_t *setting, int
val)
{
	unsigned long flags;
	int i;
	u32 *p;

	if (!capable(CAP_SYS_ADMIN))
		return -EACCES;
	if (!(setting->rw & SETTING_WRITE))
		return -EPERM;
	if (val < setting->min || val > setting->max)
		return -EINVAL;
	if (setting->set)
		return setting->set(drive, val);
	if (ide_spin_wait_hwgroup(drive, &flags))
		return -EBUSY;
	switch (setting->data_type) {
		case TYPE_BYTE:
			*((u8 *) setting->data) = val;
			break;
		case TYPE_SHORT:
			*((u16 *) setting->data) = val;
			break;
		case TYPE_INT:
			*((u32 *) setting->data) = val;
			break;
		case TYPE_INTA:
			p = (u32 *) setting->data;
			for (i = 0; i < 1 << PARTN_BITS; i++, p++)
				*p = val;
			break;
	}
	spin_unlock_irqrestore(&HWGROUP(drive)->spinlock, flags);
	return 0;
}

int ide_read_setting(ide_drive_t *drive, ide_settings_t *setting)
{
	int		val = -EINVAL;
	unsigned long	flags;

	if ((setting->rw & SETTING_READ)) {
		spin_lock_irqsave(&HWGROUP(drive)->spinlock, flags);
		switch(setting->data_type) {
			case TYPE_BYTE:
				val = *((u8 *) setting->data);
				break;
			case TYPE_SHORT:
				val = *((u16 *) setting->data);
				break;
			case TYPE_INT:
			case TYPE_INTA:
				val = *((u32 *) setting->data);
				break;
		}
		spin_unlock_irqrestore(&HWGROUP(drive)->spinlock, flags);
	}
	return val;
}

#endif /* _IDE_I_H */

